/*
 * FILE:
 *   xna_gfx_context.cpp
 *
 * PURPOSE:
 *   Implement GfxContext for Mips/XNA
 *
 * AUTHOR:
 *   Stephen Thompson <stephen@solarflare.org.uk>
 *
 * CREATED:
 *   19-Jun-2011
 *   
 * COPYRIGHT:
 *   Copyright (C) Stephen Thompson, 2011.
 *
 *   This file is part of Mips2cs. Mips2cs is distributed under the terms
 *   of the Boost Software License, Version 1.0, the text of which
 *   appears below.
 *
 *   Boost Software License - Version 1.0 - August 17th, 2003
 *   
 *   Permission is hereby granted, free of charge, to any person or organization
 *   obtaining a copy of the software and accompanying documentation covered by
 *   this license (the "Software") to use, reproduce, display, distribute,
 *   execute, and transmit the Software, and to prepare derivative works of the
 *   Software, and to permit third-parties to whom the Software is furnished to
 *   do so, all subject to the following:
 *
 *   The copyright notices in the Software and this entire statement, including
 *   the above license grant, this restriction and the following disclaimer,
 *   must be included in all copies of the Software, in whole or in part, and
 *   all derivative works of the Software, unless such copies or derivative
 *   works are solely in the form of machine-executable object code generated by
 *   a source language processor.
 *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 *   SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 *   FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *   DEALINGS IN THE SOFTWARE.
 * 
 */

#include "xna_gfx_context.hpp"
#include "xna_graphic.hpp"
#include "encode_color.hpp"

#include "vm_calls.h"

Coercri::XNAGfxContext::~XNAGfxContext()
{
    vm_end_batches();
    vm_yield(); // C# will then flip the buffers
}

void Coercri::XNAGfxContext::setClipRectangle(const Rectangle &rect)
{
    vm_set_clip_rectangle(rect.getLeft(), rect.getTop(), rect.getWidth(), rect.getHeight());
}

void Coercri::XNAGfxContext::clearClipRectangle()
{
    vm_clear_clip_rectangle();
}

Coercri::Rectangle Coercri::XNAGfxContext::getClipRectangle() const
{
    int x, y, w, h;
    vm_get_clip_rectangle(&x, &y, &w, &h);
    return Rectangle(x, y, w, h);
}

int Coercri::XNAGfxContext::getWidth() const
{
    int w, h;
    vm_get_window_size(&w, &h);
    return w;
}

int Coercri::XNAGfxContext::getHeight() const
{
    int w, h;
    vm_get_window_size(&w, &h);
    return h;
}

void Coercri::XNAGfxContext::clearScreen(Color col)
{
    vm_clear_screen(EncodeColor(col));
}

void Coercri::XNAGfxContext::plotPixel(int x, int y, Color col)
{
    vm_plot_pixel(x, y, EncodeColor(col));
}

void Coercri::XNAGfxContext::drawGraphic(int x, int y, const Graphic &graphic)
{
    vm_draw_graphic(x, y, dynamic_cast<const XNAGraphic&>(graphic).getID());
}

void Coercri::XNAGfxContext::drawText(int x, int y, const Font &font, const std::string &text, Color col, bool antialias)
{
    // TODO
}

void Coercri::XNAGfxContext::drawLine(int x1, int y1, int x2, int y2, Color col)
{
    struct { int x1, y1, x2, y2, col; } args;
    args.x1 = x1;
    args.y1 = y1;
    args.x2 = x2;
    args.y2 = y2;
    args.col = EncodeColor(col);
    vm_draw_line(&args);
}

void Coercri::XNAGfxContext::fillRectangle(const Rectangle &rect, Color col)
{
    struct { int x1, y1, x2, y2, col; } args;
    args.x1 = rect.getLeft();
    args.y1 = rect.getTop();
    args.x2 = rect.getRight();
    args.y2 = rect.getBottom();
    args.col = EncodeColor(col);
    vm_fill_rectangle(&args);
}
