/*
 * FILE:
 *   xna_gfx_driver.cpp
 *
 * PURPOSE:
 *   XNA/MIPS implementation of GfxDriver
 *
 * AUTHOR:
 *   Stephen Thompson <stephen@solarflare.org.uk>
 *
 * CREATED:
 *   18-Jun-2011
 *   
 * COPYRIGHT:
 *   Copyright (C) Stephen Thompson, 2011.
 *
 *   This file is part of Mips2cs. Mips2cs is distributed under the terms
 *   of the Boost Software License, Version 1.0, the text of which
 *   appears below.
 *
 *   Boost Software License - Version 1.0 - August 17th, 2003
 *   
 *   Permission is hereby granted, free of charge, to any person or organization
 *   obtaining a copy of the software and accompanying documentation covered by
 *   this license (the "Software") to use, reproduce, display, distribute,
 *   execute, and transmit the Software, and to prepare derivative works of the
 *   Software, and to permit third-parties to whom the Software is furnished to
 *   do so, all subject to the following:
 *
 *   The copyright notices in the Software and this entire statement, including
 *   the above license grant, this restriction and the following disclaimer,
 *   must be included in all copies of the Software, in whole or in part, and
 *   all derivative works of the Software, unless such copies or derivative
 *   works are solely in the form of machine-executable object code generated by
 *   a source language processor.
 *
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 *   SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 *   FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 *   ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *   DEALINGS IN THE SOFTWARE.
 * 
 */

#include "xna_gfx_driver.hpp"
#include "xna_graphic.hpp"
#include "xna_window.hpp"

#include "gfx/pixel_array.hpp"
#include "gfx/window_listener.hpp"

#include "vm_calls.h"

#include <string>

Coercri::GfxDriver::DisplayModeVector Coercri::XNAGfxDriver::getFullScreenModes()
{
    int num = vm_get_num_full_screen_modes();
    DisplayModeVector result(num);
    vm_get_full_screen_modes(&result[0]);
    return result;
}

Coercri::GfxDriver::DisplayMode Coercri::XNAGfxDriver::getDesktopMode()
{
    DisplayMode dm;
    vm_get_desktop_mode(&dm);
    return dm;
}

boost::shared_ptr<Coercri::Window> Coercri::XNAGfxDriver::createWindow(int width, int height,
								       bool resizable, bool fullscreen,
								       const std::string &title)
{
    struct { int w, h, resiz, full; const char * title; } args;
    args.w = width;
    args.h = height;
    args.resiz = resizable ? 1 : 0;
    args.full = fullscreen ? 1 : 0;
    args.title = title.c_str();
    vm_create_window(&args);
    window.reset(new XNAWindow);
    return window;
}

boost::shared_ptr<Coercri::Graphic> Coercri::XNAGfxDriver::createGraphic(boost::shared_ptr<const PixelArray> pixels, int hx, int hy)
{
    struct { const void * pix; int w, h, hx, hy; } args;
    args.pix = &(*pixels)(0,0);
    args.w = pixels->getWidth();
    args.h = pixels->getHeight();
    args.hx = hx;
    args.hy = hy;
    int id = vm_create_graphic(&args);
    return boost::shared_ptr<Graphic>(new XNAGraphic(id));
}

bool Coercri::XNAGfxDriver::pollEvents()
{
    bool did_something = false;
    typedef std::vector<WindowListener*> wl_vec;
    typedef wl_vec::const_iterator iter;
    if (window) {
	const wl_vec & wls = window->getListeners();

        // mouse move event
	int new_mx, new_my;
	vm_get_mouse_pos(&new_mx, &new_my);
	
	if (new_mx != mx || new_my != my) {
	    mx = new_mx;
	    my = new_my;
	    for (iter it = wls.begin(); it != wls.end(); ++it) {
		(*it)->onMouseMove(mx, my);
	    }
	    did_something = true;
	}
    }
    
    return did_something;
}

void Coercri::XNAGfxDriver::setKeyRepeat(bool enabled)
{
    // Not implemented
}
